# -*- coding: utf-8 -*-

"""Project wide utils related to `subprocess` module."""

from __future__ import annotations

__all__ = ['CREATE_NO_WINDOW']

import sys
import subprocess  # nosec
import typing


CREATE_NO_WINDOW: typing.Final[int] = subprocess.CREATE_NO_WINDOW if (sys.platform == 'win32') else 0  # type: ignore



# -*- coding: utf-8 -*-

# pylint: disable=protected-access

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Components for environment management."""

__all__ = ['ApplicationsComponent']

import html
import os
import typing
import webbrowser
from qtpy import QtCore
from anaconda_navigator.api import external_apps
from anaconda_navigator import config as anaconda_config
from anaconda_navigator.utils import constants
from anaconda_navigator.utils import launch
from anaconda_navigator.utils import telemetry
from anaconda_navigator.utils import version_utils
from anaconda_navigator.widgets import dialogs
from anaconda_navigator.widgets.dialogs import environment as environment_dialogs
from . import common

if typing.TYPE_CHECKING:
    from anaconda_navigator.widgets import main_window


class ApplicationsComponent(common.Component):
    """Component for launching third-party applications."""

    __alias__ = 'applications'

    def __init__(self, parent: 'main_window.MainWindow') -> None:
        """Initialize new :class:`~EnvironmentsComponent` instance."""
        super().__init__(parent=parent)

        self.__running_processes: typing.Final[typing.List[launch.RunningProcess]] = []

        self.__listener_timer: typing.Final[QtCore.QTimer] = QtCore.QTimer()
        self.__listener_timer.timeout.connect(self.update_running_processes)
        self.__listener_timer.setInterval(5000)
        self.__listener_timer.start()

        self.__feedback_timer: typing.Final[QtCore.QTimer] = QtCore.QTimer()
        self.__feedback_timer.timeout.connect(self.__feedback_timeout)
        self.__feedback_timer.setSingleShot(True)
        self.__feedback_timer.setInterval(5000)

    @property
    def running_processes(self) -> typing.List[launch.RunningProcess]:
        """Collection of currently launched processes."""
        return self.__running_processes[:]

    def update_running_processes(self) -> None:
        """Update status of applications launched from Navigator."""
        index: int
        for index in reversed(range(len(self.__running_processes))):
            running_application: launch.RunningProcess = self.__running_processes[index]
            if running_application.return_code is None:
                continue

            if (running_application.return_code != 0) and (running_application.age.total_seconds() < 15.0):
                self.show_application_launch_errors(running_application)

            del self.__running_processes[index]
            running_application.cleanup()

    def launch_application(  # pylint: disable=too-many-arguments
            self,
            package_name: str,
            command: str,
            extra_arguments: typing.Iterable[typing.Any],
            leave_path_alone: bool,
            prefix: str,
            sender: str,  # pylint: disable=unused-argument
            non_conda: bool,
            app_type: constants.AppType,
    ) -> None:
        """
        Launch application from home screen.

        :param package_name: Name of the conda package, or alias of the external application.
        :param command: Exact command to launch application with.
        :param extra_arguments: Additional arguments to attach to command.
        :param prefix: Conda prefix, which should be active.
        :param app_type: Type of the application.
        """
        self.main_window.update_status(action=f'Launching <b>{package_name}</b>', value=0, max_value=0)

        def next_step(*args: typing.Any) -> None:  # pylint: disable=unused-argument
            self.__launch_application(
                package_name=package_name,
                command=command,
                extra_arguments=extra_arguments,
                leave_path_alone=leave_path_alone,
                prefix=prefix,
                non_conda=non_conda,
            )

        if app_type == constants.AppType.CONDA:
            next_step()

        elif app_type == constants.AppType.INSTALLABLE:
            app = external_apps.get_applications(cached=True).installable_apps[package_name]

            app.update_config(self.main_window.current_prefix)

            # Install extensions first!
            worker = app.install_extensions()  # pylint: disable=assignment-from-no-return
            worker.sig_finished.connect(next_step)
            worker.start()

        elif app_type == constants.AppType.WEB:
            webbrowser.open_new_tab(command)

            self.__feedback_timer.start()

    def __feedback_timeout(self) -> None:
        """
        Hide application checking progress bar.

        Actual feedback is done in :meth:`~ApplicationsComponent.update_running_applications`.
        """
        self.main_window.update_status()

    def __launch_application(  # pylint: disable=too-many-arguments
        self,
        package_name: str,
        command: str,
        extra_arguments: typing.Iterable[typing.Any],
        leave_path_alone: bool,
        prefix: str,
        non_conda: bool,
    ) -> None:
        """Second phase of the :meth:`~ApplicationComponent.launch_application`."""
        environment: typing.Dict[str, str] = dict(os.environ)
        environment.pop('QT_API')

        if anaconda_config.MAC:
            # See https://github.com/ContinuumIO/anaconda-issues/issues/3287
            os.environ['LANG'] = os.environ.get('LANG') or os.environ.get('LC_ALL') or 'en_US.UTF-8'
            os.environ['LC_ALL'] = os.environ.get('LC_ALL') or os.environ['LANG']

            # See https://github.com/ContinuumIO/navigator/issues/1233
            environment['EVENT_NOKQUEUE'] = '1'

        running_process: typing.Optional[launch.RunningProcess] = launch.launch(
            root_prefix=self.main_window.api.ROOT_PREFIX,
            prefix=prefix,
            command=command,
            extra_arguments=extra_arguments,
            package_name=package_name,
            environment=environment,
            leave_path_alone=leave_path_alone,
            non_conda=non_conda,
        )
        if running_process is None:
            return

        self.__running_processes.append(running_process)

        self.__feedback_timer.start()

    def show_application_launch_errors(self, application: launch.RunningProcess) -> None:
        """Show a dialog with details on application launch error."""
        self.main_window.update_status()
        if not self.main_window.config.get('main', 'show_application_launch_errors'):
            return

        content: typing.Optional[str] = application.stderr
        if content:
            content = content.strip()
        else:
            content = f'Exit code: {application.return_code}'

        telemetry.ANALYTICS.instance.event('navigate', {'location': '/home/launch_error'})
        dialogs.MessageBoxError(
            text=f'Application <b>{application.package}</b> launch may have produced errors.',
            title='Application launch error',
            error=content,
            report=False,
            learn_more=None,
        ).exec_()
        telemetry.ANALYTICS.instance.event('navigate', {'location': '/home'})

    def check_dependencies_before_install(  # pylint: disable=too-many-statements
            self, worker, output, error,  # pylint: disable=unused-argument
    ):
        """
        Check if the package to be installed changes navigator dependencies.

        This check is made for Orange3 which is not qt5 compatible.
        """
        if isinstance(output, dict):
            exception_type = str(output.get('exception_type', ''))
            actions = output.get('actions', {})
        else:
            exception_type = ''
            actions = {}

        conflicts = False
        nav_deps_conflict = self.main_window.api.check_navigator_dependencies(actions, self.main_window.current_prefix)
        conflict_message = ''

        # Try to install in a new environment
        if 'UnsatisfiableError' in exception_type or nav_deps_conflict:
            conflicts = True
            # Try to set the default python to None to avoid issues that
            # prevent a package to be installed in a new environment due to
            # python pinning, fusion for 2.7, rstudio on win for 2.7 etc.
            self.main_window.api.conda_config_set('default_python', None)

        if conflicts:
            telemetry.ANALYTICS.instance.event('navigate', {'location': '/conflict_creating_environment'})
            dlg = environment_dialogs.ConflictDialog(
                parent=self.main_window,
                package=worker.pkgs[0],
                extra_message=conflict_message,
                current_prefix=self.main_window.current_prefix,
            )
            self.main_window._dialog_environment_action = dlg
            worker_info = self.main_window.api.conda_data(prefix=self.main_window.current_prefix)
            worker_info.sig_chain_finished.connect(dlg.setup)

            if dlg.exec_():
                env_prefix = dlg.prefix
                action_msg = f'Installing application <b>{worker.pkgs[0]}</b> on newenvironment <b>{env_prefix}</b>'

                if env_prefix not in dlg.environments:
                    new_worker = self.main_window.api.create_environment(
                        prefix=env_prefix,
                        packages=worker.pkgs,
                        no_default_python=True,
                    )
                    # Save the old prefix in case of errors
                    new_worker.old_prefix = worker.prefix

                    new_worker.action_msg = action_msg
                    new_worker.sig_finished.connect(self.main_window._conda_output_ready)
                    new_worker.sig_partial.connect(self.main_window._conda_partial_output_ready)
                else:
                    new_worker = self.main_window.api.install_packages(
                        prefix=env_prefix,
                        pkgs=worker.pkgs,
                        no_default_python=True,
                    )
                    # Save the old prefix in case of errors
                    new_worker.old_prefix = worker.prefix

                    new_worker.action = constants.ACTION_INSTALL
                    new_worker.action_msg = action_msg
                    new_worker.pkgs = worker.pkgs
                    new_worker.sig_finished.connect(self.main_window._conda_output_ready)
                    new_worker.sig_partial.connect(self.main_window._conda_partial_output_ready)
                self.main_window.update_status(action_msg, value=0, max_value=0)
            else:
                self.main_window.set_widgets_enabled(True)
                self.main_window.set_busy_status(conda=False)
                self.main_window.update_status()

            self.main_window._dialog_environment_action = None
            telemetry.ANALYTICS.instance.event('navigate', {'location': '/environments'})
        else:
            if worker.action == constants.APPLICATION_INSTALL:
                action_msg = f'Install application <b>{worker.pkgs[0]}</b> on <b>{worker.prefix}</b>'
            elif worker.action == constants.APPLICATION_UPDATE:
                action_msg = f'Updating application <b>{worker.pkgs[0]}</b> on <b>{worker.prefix}</b>'
            new_worker = self.main_window.api.install_packages(
                prefix=worker.prefix,
                pkgs=worker.pkgs,
            )
            new_worker.action_msg = action_msg
            new_worker.action = worker.action
            new_worker.sender = worker.sender
            new_worker.non_conda = worker.non_conda
            new_worker.pkgs = worker.pkgs
            new_worker.sig_finished.connect(self.main_window._conda_output_ready)
            new_worker.sig_partial.connect(self.main_window._conda_partial_output_ready)
            self.main_window.update_status(action_msg, value=0, max_value=0)

    def check_license_requirements(self, worker, output, error):
        """Check if package requires licensing and try to get a trial."""
        worker.output = output
        self.check_dependencies_before_install(worker, output, error)

    def conda_application_action(  # pylint: disable=missing-function-docstring,too-many-arguments,too-many-statements
            self, action, package_name, version, sender, non_conda, app_type,
    ):
        if app_type == constants.AppType.INSTALLABLE and action == constants.APPLICATION_INSTALL:
            self.install_external_app(package_name)
            return

        self.main_window.tab_home.set_widgets_enabled(False)
        if 'environments' in self.main_window.components:
            self.main_window.components.environments.tab.set_widgets_enabled(False)
        self.main_window.set_busy_status(conda=True)
        current_version = self.main_window.api.conda_package_version(
            pkg=package_name,
            prefix=self.main_window.current_prefix,
        )

        if version:
            pkgs = [f'{package_name}=={version}']
        else:
            pkgs = [f'{package_name}']

        if action == constants.APPLICATION_INSTALL:
            worker = self.main_window.api.install_packages(
                prefix=self.main_window.current_prefix,
                pkgs=pkgs,
                dry_run=True,
            )
            text_action: str = 'Installing'

            if current_version:
                text_action = {
                    -1: 'Downgrading',
                    1: 'Upgrading',
                }.get(
                    version_utils.compare(version, current_version),
                    text_action,
                )

            action_msg = (
                f'{html.escape(text_action)} application <b>{html.escape(package_name)}</b> on '
                f'<b>{html.escape(self.main_window.current_prefix)}</b>'
            )
            worker.prefix = self.main_window.current_prefix
            worker.action = action
            worker.action_msg = action_msg
            worker.sender = sender
            worker.pkgs = pkgs
            worker.non_conda = non_conda
            worker.sig_finished.connect(self.check_license_requirements)
            worker.sig_partial.connect(self.main_window._conda_partial_output_ready)
        elif action == constants.APPLICATION_UPDATE:
            worker = self.main_window.api.install_packages(
                prefix=self.main_window.current_prefix,
                pkgs=pkgs,
                dry_run=True,
            )
            action_msg = (
                f'Updating application <b>{html.escape(package_name)}</b> on '
                f'<b>{html.escape(self.main_window.current_prefix)}</b>'
            )
            worker.prefix = self.main_window.current_prefix
            worker.action = action
            worker.action_msg = action_msg
            worker.sender = sender
            worker.pkgs = pkgs
            worker.non_conda = non_conda
            worker.sig_finished.connect(self.check_license_requirements)
            worker.sig_partial.connect(self.main_window._conda_partial_output_ready)
        elif action == constants.APPLICATION_REMOVE:
            worker = self.main_window.api.remove_packages(prefix=self.main_window.current_prefix, pkgs=pkgs)
            action_msg = (
                f'Removing application <b>{html.escape(package_name)}</b> from '
                f'<b>{html.escape(self.main_window.current_prefix)}</b>'
            )
            worker.action = action
            worker.action_msg = action_msg
            worker.sender = sender
            worker.pkgs = pkgs
            worker.non_conda = non_conda
            worker.sig_finished.connect(self.main_window._conda_output_ready)
            worker.sig_partial.connect(self.main_window._conda_partial_output_ready)
        self.main_window.update_status(action_msg, value=0, max_value=0)

    @staticmethod
    def install_external_app(app_name):
        """Installing external app (VSCode, pycharm)."""
        external_apps.get_applications(cached=True).installable_apps[app_name].install()

    def setup(self, worker: typing.Any, output: typing.Any, error: str, initial: bool) -> None:
        """Perform component configuration from `conda_data`."""
        prefix: str
        for prefix in output['processed_info']['__environments']:
            launch.remove_package_logs(root_prefix=self.main_window.api.ROOT_PREFIX, prefix=prefix)



# -*- coding: utf-8 -*-

"""Constants for third-party applications."""

__all__ = ['EMPTY_PATCH', 'MANUAL_PATCHES']

import typing
from anaconda_navigator.static import images
from . import base


EMPTY_PATCH: typing.Final[base.AppPatch] = base.AppPatch()
MANUAL_PATCHES: typing.Final[typing.Mapping[str, base.AppPatch]] = {
    'anaconda-fusion': base.AppPatch(
        description=(
            'Integration between Excel ® and Anaconda via Notebooks. Run data science functions, interact with results '
            'and create advanced visualizations in a code-free app inside Excel.'
        ),
    ),
    'anaconda-mosaic': base.AppPatch(
        description=(
            'Interactive exploration of larger than memory datasets. Create data sources, perform transformations and '
            'combinations.'
        ),
    ),
    'anacondafusion': base.AppPatch(
        is_available=False,
    ),
    'console_shortcut': base.AppPatch(
        description='Run a cmd.exe terminal with your current environment from Navigator activated',
        display_name='CMD.exe Prompt',
    ),
    'glueviz': base.AppPatch(
        display_name='Glueviz',
        description=(
            'Multidimensional data visualization across files. Explore relationships within and among related datasets.'
        ),
        image_path=images.GLUEVIZ_ICON_1024_PATH,
    ),
    'ipython-notebook': base.AppPatch(
        image_path=images.NOTEBOOK_ICON_1024_PATH,
        is_available=False,
    ),
    'ipython-qtconsole': base.AppPatch(
        image_path=images.QTCONSOLE_ICON_1024_PATH,
        is_available=False,
    ),
    'jupyterlab': base.AppPatch(
        display_name='JupyterLab',
        description=(
            'An extensible environment for interactive and reproducible computing, based on the Jupyter Notebook and '
            'Architecture.'
        ),
        image_path=images.JUPYTERLAB_ICON_1024_PATH,
    ),
    'notebook': base.AppPatch(
        display_name='Notebook',
        description=(
            'Web-based, interactive computing notebook environment. Edit and run human-readable docs while describing '
            'the data analysis.'
        ),
        image_path=images.NOTEBOOK_ICON_1024_PATH,
    ),
    'orange-app': base.AppPatch(
        description=(
            'Component based data mining framework. Data visualization and data analysis for novice and expert. '
            'Interactive workflows with a large toolbox.'
        ),
        image_path=images.ORANGE_ICON_1024_PATH,
    ),
    'orange3': base.AppPatch(
        display_name='Orange 3',
        description=(
            'Component based data mining framework. Data visualization and data analysis for novice and expert. '
            'Interactive workflows with a large toolbox.'
        ),
        image_path=images.ORANGE_ICON_1024_PATH,
    ),
    'powershell_shortcut': base.AppPatch(
        description='Run a Powershell terminal with your current environment from Navigator activated',
        display_name='Powershell Prompt',
    ),
    'pyvscode': base.AppPatch(
        image_path=images.VSCODE_ICON_1024_PATH,
    ),
    'qt3dstudio': base.AppPatch(
        description=(
            'Rapidly build and prototype high quality 2D and 3D user interfaces using the built-in material and '
            'effects library or import your own design assets.'
        ),
        image_path=images.QTCREATOR_ICON_1024_PATH,
    ),
    'qtconsole': base.AppPatch(
        display_name='Qt Console',
        description=(
            'PyQt GUI that supports inline figures, proper multiline editing with syntax highlighting, graphical '
            'calltips, and more.'
        ),
        image_path=images.QTCONSOLE_ICON_1024_PATH,
    ),
    'qtcreator': base.AppPatch(
        description='Cross platform integrated development environment (IDE) to create C++ and QML applications.',
        image_path=images.QTCREATOR_ICON_1024_PATH,
    ),
    'rodeo': base.AppPatch(
        description=(
            'A browser-based IDE for data science with python. Includes autocomplete, syntax highlighting, IPython '
            'support.'
        ),
        image_path=images.RODEO_ICON_1024_PATH,
    ),
    'rstudio': base.AppPatch(
        display_name='RStudio',
        description=(
            'A set of integrated tools designed to help you be more productive with R. Includes R essentials and '
            'notebooks.'
        ),
        image_path=images.RSTUDIO_ICON_1024_PATH,
    ),
    'spyder': base.AppPatch(
        display_name='Spyder',
        description=(
            'Scientific PYthon Development EnviRonment. Powerful Python IDE with advanced editing, interactive '
            'testing, debugging and introspection features'
        ),
        image_path=images.SPYDER_ICON_1024_PATH,
    ),
    'spyder-app': base.AppPatch(
        image_path=images.SPYDER_ICON_1024_PATH,
        is_available=False,
    ),
    'veusz': base.AppPatch(
        description=(
            'Veusz is a GUI scientific plotting and graphing package. It is designed to produce publication-ready '
            'Postscript or PDF output.'
        ),
        image_path=images.VEUSZ_ICON_1024_PATH,
    ),
}



# -*- coding: utf-8 -*-

"""Additional types used by CloudAPI."""

from __future__ import annotations

__all__ = ()

import typing


class CloudEnvironmentRecord(typing.TypedDict):
    """Description of the environment reported by Cloud API."""

    id: str  # pylint: disable=invalid-name
    name: str
    yaml_ref: str
    revision: int
    created_at: str
    updated_at: str


class CloudEnvironmentCollection(typing.TypedDict):
    """Collection of environments reported by Cloud API."""

    items: typing.Sequence[CloudEnvironmentRecord]
    total: int



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Update application dialog."""

from qtpy.QtWidgets import QHBoxLayout, QLabel, QVBoxLayout  # pylint: disable=no-name-in-module
from anaconda_navigator.config import CONF
from anaconda_navigator.widgets import ButtonPrimary, QCheckBox, SpacerHorizontal, SpacerVertical
from anaconda_navigator.widgets.dialogs import DialogBase


class DialogOfflineMode(DialogBase):
    """Offline mode dialog."""

    WIDTH = 460
    _MESSAGE_BASE = (
        'Some of the functionality of Anaconda Navigator will be limited. '  # pylint: disable=implicit-str-concat
        'Conda environment creation will be subject to the packages currently available on your package '
        'cache.<br><br>'
    )
    _MESSAGE_LOC = (
        '<b>Offline mode</b> is indicated to the left of the login/logout '  # pylint: disable=implicit-str-concat
        'button on the top right corner of the main application window.<br><br>'
    )
    _MESSAGE_ENABLE = 'Offline mode will be disabled automatically when internet connectivity is restored.<br><br>'
    _MESSAGE_FORCE = (
        'You can also manually force <b>Offline mode</b> by enabling the setting on the application preferences.<br>'
    )
    _MESSAGE_EXTRA_PREF = 'By checking this option you will force <b>Offline mode</b>.<br>'

    MESSAGE_TOOL = _MESSAGE_BASE + _MESSAGE_ENABLE + _MESSAGE_FORCE
    MESSAGE_PREFERENCES = _MESSAGE_BASE + _MESSAGE_LOC + _MESSAGE_ENABLE + _MESSAGE_EXTRA_PREF
    MESSAGE_DIALOG = _MESSAGE_BASE + _MESSAGE_LOC + _MESSAGE_ENABLE + _MESSAGE_FORCE

    def __init__(
        self,
        parent=None,
        config=CONF,
    ):
        """Offline mode dialog."""
        super().__init__(parent=parent)

        self.label = QLabel(self.MESSAGE_DIALOG)
        self.button_ok = ButtonPrimary('Ok')
        self.checkbox_hide = QCheckBox("Don't show again")
        self.config = config

        # Widgets setup
        self.label.setWordWrap(True)
        self.setMinimumWidth(self.WIDTH)
        self.setMaximumWidth(self.WIDTH)
        self.setWindowTitle('Offline Mode')

        # Layouts
        layout_buttons = QHBoxLayout()
        layout_buttons.addStretch()
        layout_buttons.addWidget(self.checkbox_hide)
        layout_buttons.addWidget(SpacerHorizontal())
        layout_buttons.addWidget(self.button_ok)

        layout = QVBoxLayout()
        layout.addWidget(self.label)
        layout_buttons.addWidget(SpacerVertical())
        layout_buttons.addWidget(SpacerVertical())
        layout.addLayout(layout_buttons)
        self.setLayout(layout)

        # Signals
        self.button_ok.clicked.connect(self.handle_accept)

        # Setup
        self.button_ok.setFocus()
        self.setup()

    def setup(self):
        """Setup widget content."""
        hide_dialog = self.config.get('main', 'hide_offline_dialog')
        self.checkbox_hide.setChecked(bool(hide_dialog))

    def handle_accept(self):
        """Handle not showing updates on startup."""
        value = bool(self.checkbox_hide.checkState())
        self.config.set('main', 'hide_offline_dialog', value)
        self.accept()
